/* @copyright Itential, LLC 2020 */

/* eslint no-console: warn */
/* eslint import/no-unresolved: warn */
/* eslint global-require: warn */

// suppress eslint rule in adapter
/* eslint arrow-parens: warn */
/* eslint import/no-extraneous-dependencies: warn */
/* eslint import/no-dynamic-require: warn */

const path = require('path');
const program = require('commander');
const rls = require('readline-sync');
const fs = require('fs-extra');

const utils = require('./tbUtils');
const basicGet = require('./basicGet');
const { name } = require('../package.json');
const sampleProperties = require('../sampleProperties.json');
const adapterPronghorn = require('../pronghorn.json');

// send interactive questions and collection answers
// return updated connection object
const collectAnswersSync = (questions, props) => {
  const answers = [];
  questions.forEach(q => {
    const answer = rls.question(q);
    answers.push(answer);
  });
  return utils.getNewProps(answers, props);
};

// change object into array of questions
const confirm = (props) => {
  const questions = Object.keys(props).map(key => `${key}: (${props[key]}) `);
  return collectAnswersSync(questions, props);
};

// allow user to change auth_method
const confirmAuthOptions = (authentication) => {
  const authOptions = ['basic user_password', 'request_token', 'static_token', 'no_authentication'];
  const displayAuthOptions = utils.getDisplayAuthOptions(authentication.auth_method, authOptions);
  const index = rls.keyInSelect(displayAuthOptions, 'Which authentication?');
  if (index === -1) {
    return authentication.auth_method;
  }
  console.log(`${authOptions[index]} is selected.`);
  return authOptions[index];
};

// helper function to update auth properties
const confirmAndUpdate = (auth, config) => {
  const newAuth = confirm(auth);
  return utils.updateAuth(newAuth, auth, config);
};

// extract basic auth properties
const updateBasicAuth = (config, authentication) => {
  const auth = {
    username: authentication.username,
    password: authentication.password
  };
  return confirmAndUpdate(auth, config);
};

// extract static auth properties
const updateStaticAuth = (config, authentication) => {
  const auth = {
    token: authentication.token,
    auth_field: authentication.auth_field,
    auth_field_format: authentication.auth_field_format
  };
  return confirmAndUpdate(auth, config);
};

// troubleshooting connection and healthcheck endpoint setting of adapter
const VerifyHealthCheckEndpoint = (config) => {
  const connection = utils.getConnection(config.properties);
  const newConnection = confirm(connection);
  console.log(newConnection);
  utils.runConnectivity(newConnection.host);
  const connConfig = utils.updateNewConnection(config, newConnection);

  const healthcheck = require('../entities/.system/action.json');
  const healthCheckEndpoint = utils.getHealthCheckEndpoint(healthcheck);
  const newHealthCheckEndpoint = confirm(healthCheckEndpoint);
  utils.getHealthCheckEndpointURL(newHealthCheckEndpoint, connConfig);

  const { authentication } = connConfig.properties.properties;
  authentication.auth_method = confirmAuthOptions(authentication);
  let authConfig = connConfig;
  if (authentication.auth_method === 'basic user_password') {
    authConfig = updateBasicAuth(connConfig, authentication);
  } else if (authentication.auth_method === 'static_token') {
    authConfig = updateStaticAuth(connConfig, authentication);
  } else if (authentication.auth_method === 'request_token') {
    console.log('current troubleshooting script does not support updating request_token authentication');
  }
  utils.updateHealthCheckEndpoint(newHealthCheckEndpoint, healthCheckEndpoint, healthcheck);
  return authConfig;
};

const getPronghornProps = (iapDir) => {
  console.log('Retrieving properties.json file...');
  const rawProps = require(path.join(iapDir, 'properties.json'));
  console.log('Decrypting properties...');
  const { Discovery } = require('@itential/itential-utils');
  const discovery = new Discovery();
  const pronghornProps = utils.decryptProperties(rawProps, __dirname, discovery);
  console.log('Found properties.\n');
  return pronghornProps;
};

const offline = async () => {
  console.log('Start offline troubleshooting');
  const updatedAdapter = VerifyHealthCheckEndpoint({ properties: sampleProperties });
  const a = basicGet.getAdapterInstance(updatedAdapter);
  const res = await utils.healthCheck(a);
  if (!res) {
    console.log('run `npm run troubleshoot` again to update settings');
    process.exit(0);
  }
  console.log('Save changes to sampleProperties.json');
  await fs.writeFile('sampleProperties.json', JSON.stringify(updatedAdapter.properties, null, 2));
  if (rls.keyInYN('Test with more GET request')) {
    await utils.runBasicGet();
  }
};

// Main.  Let's do this!

const main = async (command) => {
  const iapDir = path.join(__dirname, '../../../../');
  if (!utils.withinIAP(iapDir)) {
    if (command === 'connectivity') {
      const { host } = sampleProperties.properties;
      console.log(`perform networking diagnositics to ${host}`);
      await utils.runConnectivity(host);
      process.exit(0);
    } else if (command === 'healthcheck') {
      const a = basicGet.getAdapterInstance({ properties: sampleProperties });
      await utils.healthCheck(a);
      process.exit(0);
    } else if (command === 'basicget') {
      await utils.runBasicGet();
      process.exit(0);
    }
    if (rls.keyInYN('Troubleshooting without IAP?')) {
      await offline();
    }
    process.exit(0);
  }
  const pronghornProps = getPronghornProps(iapDir);
  console.log('Connecting to Database...');
  const database = await basicGet.connect(pronghornProps);
  console.log('Connection established.');
  const filter = { id: pronghornProps.id };

  const profileItem = await database.collection(utils.IAP_PROFILES_COLLECTION).findOne(filter);
  if (!profileItem) {
    console.log(`Could not find IAP profile for id ${pronghornProps.id}`);
    process.exit(0);
  }
  const serviceItem = await database.collection(utils.SERVICE_CONFIGS_COLLECTION).findOne(
    { model: name }
  );

  if (command === undefined) {
    if (serviceItem) {
      if (rls.keyInYN(`Start verifying the connection and authentication properties for ${name}?`)) {
        const adapter = serviceItem;
        const updatedAdapter = VerifyHealthCheckEndpoint(adapter);
        const a = basicGet.getAdapterInstance(updatedAdapter);
        const res = await utils.healthCheck(a);
        if (!res) {
          console.log('run `npm run troubleshoot` again to update settings');
          process.exit(0);
        }
        const update = { $set: { properties: updatedAdapter.properties } };
        await database.collection(utils.SERVICE_CONFIGS_COLLECTION).updateOne(
          { model: name }, update
        );
        console.log(`${name} updated.`);
        if (rls.keyInYN('Test with more GET request')) {
          await utils.runBasicGet();
          console.log('Restart the adapter in IAP for changes to take effect');
          process.exit(0);
        } else {
          console.log('Exiting');
          process.exit(0);
        }
      } else {
        console.log('You can update healthCheckEndpoint in ./entities/.system/action.json');
        console.log('You can update authentication credientials under Settings/Services');
        console.log('Exiting');
        process.exit(0);
      }
    } else {
      console.log(`${name} not installed`);
      console.log('run `npm run install:adapter` to install current adapter to IAP first. Exiting...');
      process.exit(0);
    }
  } else if (command === 'install') {
    if (serviceItem) {
      console.log(`A service by the name ${name} already exits!`);
      if (rls.keyInYN(`Do you want to completely remove ${name}?`)) {
        console.log(`Removing ${name} from db...`);
        await database.collection(utils.SERVICE_CONFIGS_COLLECTION).deleteOne({ model: name });
        console.log(`${name} removed from db.`);
        if (profileItem.services.includes(serviceItem.name)) {
          const serviceIndex = profileItem.services.indexOf(serviceItem.name);
          profileItem.services.splice(serviceIndex, 1);
          const update = { $set: { services: profileItem.services } };
          await database.collection(utils.IAP_PROFILES_COLLECTION).updateOne(
            { id: pronghornProps.id }, update
          );
          console.log(`${serviceItem.name} removed from profileItem.services.`);
          console.log(`Rerun the script to reinstall ${serviceItem.name}.`);
          process.exit(0);
        } else {
          process.exit(0);
        }
      } else {
        console.log('Exiting...');
        process.exit(0);
      }
    } else {
      utils.verifyInstallationDir(__dirname, name);
      utils.npmInstall();
      utils.runTest();
      if (rls.keyInYN(`Do you want to install ${name} to IAP?`)) {
        console.log('Creating database entries...');
        const adapter = utils.createAdapter(
          pronghornProps, profileItem, sampleProperties, adapterPronghorn
        );
        await database.collection(utils.SERVICE_CONFIGS_COLLECTION).insertOne(adapter);
        profileItem.services.push(adapter.name);
        const update = { $set: { services: profileItem.services } };
        await database.collection(utils.IAP_PROFILES_COLLECTION).updateOne(
          { id: pronghornProps.id }, update
        );
        console.log('Database entry creation complete.');
      }
      console.log('`npm run troubleshoot` to start troubleshooting the adapter.');
      console.log('Exiting...');
      process.exit(0);
    }
  } else if (['healthcheck', 'basicget', 'connectivity'].includes(command)) {
    if (serviceItem) {
      const adapter = serviceItem;
      const a = basicGet.getAdapterInstance(adapter);
      if (command === 'healthcheck') {
        await utils.healthCheck(a);
        process.exit(0);
      } else if (command === 'basicget') {
        await utils.runBasicGet();
      } else if (command === 'connectivity') {
        const { host } = adapter.properties.properties;
        console.log(`perform networking diagnositics to ${host}`);
        await utils.runConnectivity(host);
        process.exit(0);
      }
    } else {
      console.log(`${name} not installed. Run npm \`run install:adapter\` to install.`);
      process.exit(0);
    }
  }
};

program
  .command('connectivity')
  .alias('c')
  .description('networking diagnostics')
  .action(() => {
    main('connectivity');
  });

program
  .command('install')
  .alias('i')
  .description('install current adapter')
  .action(() => {
    main('install');
  });

program
  .command('healthcheck')
  .alias('hc')
  .description('perfom none interative healthcheck with current setting')
  .action(() => {
    main('healthcheck');
  });

program
  .command('basicget')
  .alias('bg')
  .description('perfom basicget')
  .action(() => {
    main('basicget');
  });

// Allow commander to parse `process.argv`
program.parse(process.argv);

if (process.argv.length < 3) {
  main();
}

const allowedParams = ['install', 'healthcheck', 'basicget', 'connectivity'];
if (process.argv.length === 3 && !allowedParams.includes(process.argv[2])) {
  console.log(`unknown parameter ${process.argv[2]}`);
  console.log('try `node troubleshootingAdapter.js -h` to see allowed parameters. Exiting...');
  process.exit(0);
}
